import os
import filecmp
import tempfile
import threading

from absl import flags
from absl import logging
from Bio import SeqIO

import zeta_suite
from hhblits_wrapper import hhblits
from hmmer_wrapper import hmmer
from utils import line_count

FLAGS = flags.FLAGS


def run_hhblits(infile, oa3m, db, e_value, cov, max_iter=3):
    if os.path.exists(oa3m):
        logging.info(f"{oa3m} exists, skipped")
        return
    os.makedirs(os.path.dirname(oa3m), exist_ok=True)
    with tempfile.TemporaryDirectory(dir=FLAGS.tmpdir) as tmpdirname:
        for i in range(max_iter):
            hhblits(
                infile=infile,
                oa3m=os.path.join(tmpdirname, f"{i}.a3m"),
                db=db,
                evalue=e_value,
                cov=cov,
                n_iter=1,
            )
            ofile = os.path.join(tmpdirname, f"{i}.a3m")
            if filecmp.cmp(infile, ofile):
                logging.info(
                    f"No new sequences searched at iteration {i + 1}, stopped."
                )
                break

            if line_count(ofile) > 10000:
                logging.info(f"More than 5000 sequences at iteration {i + 1}, stopped.")
                break
            infile = ofile
        zeta_suite.reorder(ofile, oa3m, top_nums=5000)


def run_hmmer(infile, oa3m, db, e_value, cov, max_iter=3):
    if os.path.exists(oa3m):
        logging.info(f"{oa3m} exists, skipped")
        return
    os.makedirs(os.path.dirname(oa3m), exist_ok=True)
    with tempfile.TemporaryDirectory(dir=FLAGS.tmpdir) as tmpdirname:
        for i in range(max_iter):
            hmmer(
                infile=infile,
                oa3m=os.path.join(tmpdirname, f"{i}.a3m"),
                db=db,
                evalue=e_value,
            )
            zeta_suite.filter(
                os.path.join(tmpdirname, f"{i}.a3m"),
                os.path.join(tmpdirname, f"{i}_f.a3m"),
                cov=cov,
            )
            ofile = os.path.join(tmpdirname, f"{i}_f.a3m")
            if filecmp.cmp(infile, ofile):
                logging.info(
                    f"No new sequences searched at iteration {i + 1}, stopped."
                )
                break

            if line_count(ofile) > 10000:
                logging.info(f"More than 5000 sequences at iteration {i + 1}, stopped.")
                break
            infile = ofile
        zeta_suite.reorder(ofile, oa3m, top_nums=5000)


def search_uniref30(fasta, outdir):
    seq = SeqIO.read(fasta, "fasta").seq
    L = len(seq)
    cov = 50#min(50, round(5000 / L))
    target_name = os.path.splitext(os.path.basename(fasta))[0]
    for e_value, suf in zip(["0.001","0.1"], ["E-3","E-1"]):
        oa3m = os.path.join(outdir, f"UniRef30_{suf}", target_name + ".a3m")
        run_hhblits(
            infile=fasta,
            oa3m=oa3m,
            db=FLAGS.uniref30_db,
            e_value=e_value,
            cov=cov,
            max_iter=3,
        )
        if line_count(oa3m) > 10000:
            logging.info(f"More than 5000 sequences at E-value {e_value}, stopped.")
            break


def search_bfd30(fasta, outdir):
    seq = SeqIO.read(fasta, "fasta").seq
    L = len(seq)
    cov = 50#min(50, round(5000 / L))
    target_name = os.path.splitext(os.path.basename(fasta))[0]
    for e_value, suf in zip(["0.001","0.1"], ["E-3","E-1"]):
        oa3m = os.path.join(outdir, f"BFD30_{suf}", target_name + ".a3m")
        run_hhblits(
            infile=fasta,
            oa3m=oa3m,
            db=FLAGS.bfd30_db,
            e_value=e_value,
            cov=cov,
            max_iter=3,
        )
        if line_count(oa3m) > 10000:
            logging.info(f"More than 5000 sequences at E-value {e_value}, stopped.")
            break


def search_uniref90(fasta, outdir):
    seq = SeqIO.read(fasta, "fasta").seq
    L = len(seq)
    cov = 50#min(50, round(5000 / L))
    target_name = os.path.splitext(os.path.basename(fasta))[0]
    for e_value, suf in zip(["0.001"], ["E-3"]):
        oa3m = os.path.join(outdir, f"UniRef90_{suf}", target_name + ".a3m")
        run_hmmer(
            infile=fasta,
            oa3m=oa3m,
            db=FLAGS.uniref90_db,
            e_value=e_value,
            cov=cov,
            max_iter=3,
        )
        if line_count(oa3m) > 10000:
            logging.info(f"More than 5000 sequences at E-value {e_value}, stopped.")
            break


def search_mgnify90(fasta, outdir):
    seq = SeqIO.read(fasta, "fasta").seq
    L = len(seq)
    cov = 50#min(50, round(5000 / L))
    target_name = os.path.splitext(os.path.basename(fasta))[0]
    for e_value, suf in zip(["0.001"], ["E-3"]):
        oa3m = os.path.join(outdir, f"MGnify90_{suf}", target_name + ".a3m")
        run_hmmer(
            infile=fasta,
            oa3m=oa3m,
            db=FLAGS.mgnify90_db,
            e_value=e_value,
            cov=cov,
            max_iter=3,
        )
        if line_count(oa3m) > 10000:
            logging.info(f"More than 5000 sequences at E-value {e_value}, stopped.")
            break


def zetahit_main():
    threads = []
    if FLAGS.use_uniref30:
        threads.append(
            threading.Thread(target=search_uniref30, args=(FLAGS.fasta, FLAGS.odir))
        )
    if FLAGS.use_uniref90:
        threads.append(
            threading.Thread(target=search_uniref90, args=(FLAGS.fasta, FLAGS.odir))
        )
    if FLAGS.use_bfd30:
        threads.append(
            threading.Thread(target=search_bfd30, args=(FLAGS.fasta, FLAGS.odir))
        )
    if FLAGS.use_mgnify90:
        threads.append(
            threading.Thread(target=search_mgnify90, args=(FLAGS.fasta, FLAGS.odir))
        )
    for x in threads:
        x.start()
    for x in threads:
        x.join()
