#!/home/konglupeng/anaconda3/bin/python
# author klp
import argparse
import numpy as np
import os
import subprocess
import sys
import shutil
os.environ['MKL_THREADING_LAYER'] = 'GNU'
from localconfig import *
from utils import *


def parse_arguments():
    parser = argparse.ArgumentParser(description="FALCON2 for Protein Structure Predicion.", usage='%(prog)s <-s sequence_name> <-i input_fasta_file> [options]' , add_help=False)
    
    input_args = parser.add_argument_group('Input arguments')
    input_args.add_argument('-s', '--seqname', metavar='<sequence_name>', help='query protein sequence name')
    input_args.add_argument('-i', '--indir', metavar='<input_dir>',  help='input directory')

    output_args = parser.add_argument_group('Output arguments')
    output_args.add_argument('-o','--outdir', metavar='<output_directory>', help='output all results to directory')
    
    general_args = parser.add_argument_group('General arguments')
    general_args.add_argument('-c', '--cpu', metavar='<number_of_cpus>', type=int, default=16, help='number of cpus, [default=16]')

    other_args = parser.add_argument_group('Other arguments')
    other_args.add_argument('-v', '--verbose', action='store_true', help='increase output verbosity, [default=False]')
    other_args.add_argument('-h', '--help', action='help', help='show this help message and exit')
    other_args.add_argument('--version', action='version', version='%(prog)s 0.1')
    
    return parser.parse_args()

def check_file(afile):
    if not os.path.exists(afile):
        sys.exit('ERROR for %s' % afile)

def check_program(program):
    try:
        with open(os.devnull, 'w') as fout:
            subprocess.call([program, '-h'], stderr=fout, stdout=fout)
    except:
        sys.exit('ERROR for %s' % program)


def check_gpu():
    try:
        with open(os.devnull, 'w') as fout:
            subprocess.call(['nvidia-smi'], stderr=fout, stdout=fout)
        return True
    except:
        return False



def main(seqname, indir, outdir, threads=4):
    print(seqname, indir, outdir, threads)
    print('Checking protein_id and input files...')
    seqfile = '%s/%s.seq' % (indir, seqname)
    check_file(seqfile)  
    #print('')

    #a3mfile = '%s/%s.a3m' % (indir, seqname)
    #check_file(a3mfile)
    #tgtfile = '%s/%s.tgt' % (indir, seqname)
    #check_file(tgtfile)
    print('input files OK.')
    
    print('Testing Dependencies...')
    check_program(ProFOLD2)
    #check_program(ProALIGN)
    check_program(buildTopModels)
    print('Dependencies OK.')

    print('Checking output dir...')
    #outdir = '%s/%s/proalign_threading' % (outdir, seqname)
    #print(outdir)
    #sys.exit('Finish.')
    check_outdir(outdir)
    print(outdir)

    print('Generating MSA...')
    cmds = [buildMSA2,seqfile,outdir]
    check_output_stdout(cmds)
    a3mfile = '%s/%s.a3m' % (outdir, seqname)
    check_file(a3mfile)
    tgtfile = '%s/%s.tgt' % (outdir, seqname)
    

    if not os.path.exists(tgtfile):
        os.chdir(TGTROOT)
        cmds = ['./A3M_To_TGT', '-i', seqfile, '-I', a3mfile, '-o', tgtfile]
        check_output_stdout(cmds) 

    #sys.exit()
    print('Predicting protein distance matrix ...')
    # convert a3m file to aln file
    alnfile = '%s/%s.aln' % (outdir, seqname)
    if not os.path.exists(alnfile):
        cmds = [a3m2aln, a3mfile, alnfile]
        check_output_stdout(cmds)


    print('Running PROFOLD...')
    pred_disfile = '%s/%s.npz' % (outdir, seqname)
    rankfile2 = '%s/%s' % (outdir, 'rank.txt')
    cmds = [ProFOLD2, alnfile, outdir]
    #cmds = [ProFOLD, '-m', PROFMODELG, '-i', alnfile, '-o', pred_disfile, '-c', '0']
    if not os.path.exists(pred_disfile):
        check_output_stdout(cmds)
    if not os.path.exists(rankfile2):
        check_output_stdout(cmds)
    rankfile2 = '%s/%s' % (outdir, 'rank.txt')
    fm_rankfile = '%s/%s_fm.rank' % (outdir, seqname)
    cmds = [getRankDis,seqname,rankfile2,outdir,str(5)]
    if not os.path.exists(fm_rankfile):
        check_output_stdout(cmds)

    dis_mat = np.load(pred_disfile)['cbcb']
    con_mat = np.sum(dis_mat[:,:,:13],-1)
    epad_prob = '%s/%s.epad_prob' % (outdir, seqname)
    with open(epad_prob, 'w') as fout:
        for i in range(con_mat.shape[0]):
            for j in range(i + 6, con_mat.shape[0]):
                prob = con_mat[i,j] * 100
                print('%d %d %.6f %.6f' % (i, j, prob, 100-prob), file=fout)
    #print('dddddddiiiiii')
    
    #sys.exit()
    print('Checking feature file...')
    if not os.path.exists(tgtfile):
        sys.exit('ERROR: wrong tgt feature file', tgtfile)
    if not os.path.exists(pred_disfile):
        sys.exit('ERROR: wrong predict distance matrix file', pred_disfile)
    elif not os.path.exists(epad_prob):
        sys.exit('ERROR: wrong distance features')
    
    print('Checking GPU left memory space...')
    #max_left_gpu_index, max_left_mem = check_max_left_gpu()
    #if max_left_mem < 20:
    #    sys.exit('ERROR: no enough memory now in servers')
   
    # Considering the consumption of the server, here run deepconthreader instead of ProALGIN
    print('Running template-based module...')
    print('Running ProALIGN...')
    print('Searching templates...')
    cmds = [ProALIGN, seqname, tgtfile, pred_disfile, PROATempls, INMED_path, str(max_left_gpu_index)]
    #sys.exit(0)
    rankfile= ''
    if not os.path.exists(rankfile):
        check_output_stdout(cmds,False)
    sys.exit(0)
    os.system('rm -r tmp/%s*' % seqname)

    print('Building top models...')
    os.chdir(outdir)
    for fasta_item in os.listdir(INMED_path):
        if fasta_item[-6:]=='.fasta':
            

           cmds = [buildTopModels, '-i', rankfile, '-k', '10', '-d', DCT_pdb_BC100,
            '-m', modeller]
           check_output_stdout(cmds)
    os.chdir(DCTROOT)
    rankfile = '%s/%s.rank2' % (outdir, seqname)
    if not os.path.exists(rankfile):
        check_output_stdout(cmds,False)
    tm_rankfile = '%s/%s_tm.rank' % (outdir, seqname) 
    cmds = ['python',reformate_rank, seqname, tgtfile, pred_disfile, PROAMODEL, PROATempls, PROADB, rankfile, outdir]
    if not os.path.exists(tm_rankfile):
        check_output_stdout(cmds,False)
    
    print('Building predicted top models...')
    os.chdir(outdir)
    cmds = [buildTopModels, '-i', rankfile, '-k', '5', '-d', DCT_pdb_BC100,
            '-m', modeller]
    #check_output_stdout(cmds)
    print('hehe.....')
    os.system(' '.join(cmds))
    

    print('Building models from distance constraints...')

    print('Ranking predicted models by ProteinGCN')
    final_rankfile = '%s/%s_final.rank' % (outdir, seqname)
    cmds = [finalRank,seqname,outdir,outdir]
    if not os.path.exists(final_rankfile):
        check_output_stdout(cmds)

    #print('Moving predicted models...')
    #lastdir = os.path.abspath(os.path.join(outdir, ".."))
    #new_fold = outdir+'t'
    #os.makedirs(new_fold,exist_ok=True)
    #for f in os.listdir(outdir):
    #    if f.endswith('.rank') or f.endswith('.pdb') or f.endswith('.aln') or f.endswith('.npz'):
    #        file = os.path.join(outdir,f)
    #        new_file = os.path.join(new_fold,f)
    #        shutil.copyfile(file,new_file)
    #copy to upper dir (unnessary)
    #for f in os.listdir():
    #    if f.endswith('.pdb'):
    #        rawpdb = '%s/%s' % (outdir, f)
    #        newpdb = '%s/../deepconthreader-%s' % (outdir, f)
    #        cmds = ['cp', rawpdb, newpdb]
    #        check_output_stdout(cmds)
    print('Run %s finish.\n' % seqname)
    with open(outdir+'.finish','w') as w: 
        w.write('Run FALCON2 done.\n')


if __name__ == '__main__':
    # parsing arguments
    args = parse_arguments()

    # backup argument verbose, useless
    args.verbose and print('verbosity turned on')

    # check input sequence name
    if not args.seqname or len(args.seqname) > 30:
        os.system(sys.argv[0] + ' -h')
        sys.exit('\n\n\n!!!ERROR: -s (--seqname) argument missing!!!\n\n\n')
    seqname = args.seqname

    # check input file
    if not args.indir:
        os.system(sys.argv[0] + ' -h')
        sys.exit('\n\n\n!!!ERROR: -i (--input) argument missing!!!\n\n\n')
    indir = os.path.abspath(args.indir)

    # check output directory
    if not args.outdir:
        os.system(sys.argv[0] + ' -h')
        sys.exit('\n\n\n!!!ERROR: --outdir argument missing!!!\n\n\n')
    outdir = os.path.abspath(args.outdir)
    
    main(seqname, indir, outdir, args.cpu)
